@@ -1,26 +1,29 @@
-
+import contextlib
 import inspect
 import logging
 
 import wpilib
 
-from robotpy_ext.misc import PreciseDelay
+from robotpy_ext.misc import NotifierDelay
 from robotpy_ext.autonomous import AutonomousModeSelector
 
 from robotpy_ext.misc.orderedclass import OrderedClass
+from robotpy_ext.misc.annotations import get_class_annotations
 
-from networktables import NetworkTable
+from networktables import NetworkTables
 
-from .magic_tunable import setup_tunables, _TunableProperty
+from .magic_tunable import setup_tunables, _TunableProperty, collect_feedbacks
+from .magic_reset import will_reset_to
 
-__all__ = ['MagicRobot']
+__all__ = ["MagicRobot"]
 
-class MagicRobot(wpilib.SampleRobot,
-                 metaclass=OrderedClass):
+
+class MagicInjectError(ValueError):
+    pass
+
+
+class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
     """
-        .. warning:: This implementation is still being developed, and may
-                     be changed during the course of the 2016 season.
-
         Robots that use the MagicBot framework should use this as their
         base robot class. If you use this as your base, you must
         implement the following methods:
@@ -53,26 +56,26 @@ class MagicRobot(wpilib.SampleRobot,
             .. warning:: Internal API, don't override; use :meth:`createObjects` instead
         """
 
-        self._exclude_from_injection = [
-            'logger', 'members'
-        ]
+        self._exclude_from_injection = ["logger", "members"]
 
         self.__last_error_report = -10
 
         self._components = []
+        self._feedbacks = []
+        self._reset_components = []
 
         # Create the user's objects and stuff here
         self.createObjects()
 
         # Load autonomous modes
-        self._automodes = AutonomousModeSelector('autonomous')
+        self._automodes = AutonomousModeSelector("autonomous")
 
         # Next, create the robot components and wire them together
         self._create_components()
-        
-        self.__nt = NetworkTable.getTable('/robot')
-        self.__nt.putBoolean('is_simulation', self.isSimulation())
-        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())
+
+        self.__nt = NetworkTables.getTable("/robot")
+        self.__nt.putBoolean("is_simulation", self.isSimulation())
+        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
 
     def createObjects(self):
         """
@@ -119,7 +122,9 @@ class MagicRobot(wpilib.SampleRobot,
         """
         func = self.teleopPeriodic.__func__
         if not hasattr(func, "firstRun"):
-            self.logger.warn("Default MagicRobot.teleopPeriodic() method... Overload me!")
+            self.logger.warning(
+                "Default MagicRobot.teleopPeriodic() method... Overload me!"
+            )
             func.firstRun = False
 
     def disabledInit(self):
@@ -146,11 +151,41 @@ class MagicRobot(wpilib.SampleRobot,
         """
         func = self.disabledPeriodic.__func__
         if not hasattr(func, "firstRun"):
-            self.logger.warn("Default MagicRobot.disabledPeriodic() method... Overload me!")
+            self.logger.warning(
+                "Default MagicRobot.disabledPeriodic() method... Overload me!"
+            )
             func.firstRun = False
 
+    def testInit(self):
+        """Initialization code for test mode should go here.
+
+        Users should override this method for initialization code which will be
+        called each time the robot enters disabled mode.
+        """
+        pass
+
+    def testPeriodic(self):
+        """Periodic code for test mode should go here."""
+        pass
+
+    def robotPeriodic(self):
+        """
+            Periodic code for all modes should go here.
+
+            Users must override this method to utilize it
+            but it is not required.
+
+            This function gets called last in each mode.
+            You may use it for any code you need to run
+            during all modes of the robot (e.g NetworkTables updates)
+
+            The default implementation will update SmartDashboard and LiveWindow.
+        """
+        wpilib.SmartDashboard.updateValues()
+        wpilib.LiveWindow.updateValues()
+
     def onException(self, forceReport=False):
-        '''
+        """
             This function must *only* be called when an unexpected exception
             has occurred that would otherwise crash the robot code. Use this
             inside your :meth:`operatorActions` function.
@@ -180,7 +215,7 @@ class MagicRobot(wpilib.SampleRobot,
 
             :param forceReport: Always report the exception to the DS. Don't
                                 set this to True
-        '''
+        """
         # If the FMS is not attached, crash the robot program
         if not self.ds.isFMSAttached():
             raise
@@ -190,13 +225,45 @@ class MagicRobot(wpilib.SampleRobot,
         now = wpilib.Timer.getFPGATimestamp()
 
         try:
-            if forceReport or (now - self.__last_error_report) > self.error_report_interval:
+            if (
+                forceReport
+                or (now - self.__last_error_report) > self.error_report_interval
+            ):
                 wpilib.DriverStation.reportError("Unexpected exception", True)
         except:
-            pass    # ok, can't do anything here
+            pass  # ok, can't do anything here
 
         self.__last_error_report = now
 
+    @contextlib.contextmanager
+    def consumeExceptions(self, forceReport=False):
+        """
+            This returns a context manager which will consume any uncaught
+            exceptions that might otherwise crash the robot.
+
+            Example usage::
+
+                def teleopPeriodic(self):
+                    with self.consumeExceptions():
+                        if self.joystick.getTrigger():
+                            self.shooter.shoot()
+
+                    with self.consumeExceptions():
+                        if self.joystick.getRawButton(2):
+                            self.ball_intake.run()
+
+                    # and so on...
+
+            :param forceReport: Always report the exception to the DS. Don't
+                                set this to True
+
+            .. seealso:: :meth:`onException` for more details
+        """
+        try:
+            yield
+        except:
+            self.onException(forceReport=forceReport)
+
     #
     # Internal API
     #
@@ -208,19 +275,21 @@ class MagicRobot(wpilib.SampleRobot,
 
             .. warning:: Internal API, don't override
         """
-        
-        self.__nt.putString('mode', 'auto')
-        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())
+
+        self.__nt.putString("mode", "auto")
+        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
+        wpilib.LiveWindow.setEnabled(False)
 
         self._on_mode_enable_components()
 
-        self._automodes.run(self.control_loop_wait_time,
-                            self._execute_components,
-                            self.onException)
+        self._automodes.run(
+            self.control_loop_wait_time,
+            (self._execute_components, self._update_feedback, self.robotPeriodic),
+            self.onException,
+        )
 
         self._on_mode_disable_components()
 
-
     def disabled(self):
         """
             This function is called in disabled mode. You should not
@@ -229,11 +298,10 @@ class MagicRobot(wpilib.SampleRobot,
 
             .. warning:: Internal API, don't override
         """
-        
-        self.__nt.putString('mode', 'disabled')
-        ds_attached = None
 
-        delay = PreciseDelay(self.control_loop_wait_time)
+        self.__nt.putString("mode", "disabled")
+        ds_attached = None
+        wpilib.LiveWindow.setEnabled(False)
 
         self._on_mode_disable_components()
         try:
@@ -241,18 +309,20 @@ class MagicRobot(wpilib.SampleRobot,
         except:
             self.onException(forceReport=True)
 
-        while self.isDisabled():
-            
-            if ds_attached != self.ds.isDSAttached():
-                ds_attached = not ds_attached
-                self.__nt.putBoolean('is_ds_attached', ds_attached)
-            
-            try:
-                self.disabledPeriodic()
-            except:
-                self.onException()
+        with NotifierDelay(self.control_loop_wait_time) as delay:
+            while self.isDisabled():
+                if ds_attached != self.ds.isDSAttached():
+                    ds_attached = not ds_attached
+                    self.__nt.putBoolean("is_ds_attached", ds_attached)
 
-            delay.wait()
+                try:
+                    self.disabledPeriodic()
+                except:
+                    self.onException()
+
+                self._update_feedback()
+                self.robotPeriodic()
+                delay.wait()
 
     def operatorControl(self):
         """
@@ -262,13 +332,12 @@ class MagicRobot(wpilib.SampleRobot,
 
             .. warning:: Internal API, don't override
         """
-        
-        self.__nt.putString('mode', 'teleop')
+
+        self.__nt.putString("mode", "teleop")
         # don't need to update this during teleop -- presumably will switch
         # modes when ds is no longer attached
-        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())
-
-        delay = PreciseDelay(self.control_loop_wait_time)
+        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
+        wpilib.LiveWindow.setEnabled(False)
 
         # initialize things
         self._on_mode_enable_components()
@@ -278,34 +347,48 @@ class MagicRobot(wpilib.SampleRobot,
         except:
             self.onException(forceReport=True)
 
-        while self.isOperatorControl() and self.isEnabled():
-            
-            #self._update_autosend()
-            
-            try:
-                self.teleopPeriodic()
-            except:
-                self.onException()
+        with NotifierDelay(self.control_loop_wait_time) as delay:
+            while self.isOperatorControl() and self.isEnabled():
+                try:
+                    self.teleopPeriodic()
+                except:
+                    self.onException()
 
-            self._execute_components()
-            delay.wait()
+                self._execute_components()
+                self._update_feedback()
+                self.robotPeriodic()
+
+                delay.wait()
 
         self._on_mode_disable_components()
 
     def test(self):
-        '''Called when the robot is in test mode'''
-        
-        self.__nt.putString('mode', 'test')
-        self.__nt.putBoolean('is_ds_attached', self.ds.isDSAttached())
-        
-        while self.isTest() and self.isEnabled():
-            wpilib.LiveWindow.run()
-            wpilib.Timer.delay(.01)
+        """Called when the robot is in test mode"""
+
+        self.__nt.putString("mode", "test")
+        self.__nt.putBoolean("is_ds_attached", self.ds.isDSAttached())
+        wpilib.LiveWindow.setEnabled(True)
+
+        try:
+            self.testInit()
+        except:
+            self.onException(forceReport=True)
+
+        with NotifierDelay(self.control_loop_wait_time) as delay:
+            while self.isTest() and self.isEnabled():
+                try:
+                    self.testPeriodic()
+                except:
+                    self.onException()
+
+                self._update_feedback()
+                self.robotPeriodic()
+                delay.wait()
 
     def _on_mode_enable_components(self):
         # initialize things
         for component in self._components:
-            if hasattr(component, 'on_enable'):
+            if hasattr(component, "on_enable"):
                 try:
                     component.on_enable()
                 except:
@@ -314,7 +397,7 @@ class MagicRobot(wpilib.SampleRobot,
     def _on_mode_disable_components(self):
         # deinitialize things
         for component in self._components:
-            if hasattr(component, 'on_disable'):
+            if hasattr(component, "on_disable"):
                 try:
                     component.on_disable()
                 except:
@@ -331,44 +414,68 @@ class MagicRobot(wpilib.SampleRobot,
         #       other components just in case
 
         components = []
-        
+
         self.logger.info("Creating magic components")
 
         # Identify all of the types, and create them
         cls = self.__class__
 
-        for m in self.members:
-            if m.startswith('_') or isinstance(getattr(cls, m, None), _TunableProperty):
+        # - Iterate over class variables with type annotations
+        for m, ctyp in get_class_annotations(cls).items():
+            # Ignore private variables
+            if m.startswith("_"):
                 continue
-            
+
+            if hasattr(cls, m):
+                attr = getattr(cls, m)
+                # If the value given to the variable is an instance of a type and isn't a property
+                # raise an error. No double declaring types, e.g foo: type = type
+                if isinstance(attr, type) and not isinstance(attr, property):
+                    raise ValueError(
+                        "%s.%s has two type declarations" % (cls.__name__, m)
+                    )
+                # Otherwise, skip this set class variable
+                continue
+
+            # If the variable has been assigned in __init__ or createObjects, skip it
+            if hasattr(self, m):
+                continue
+
+            # If the type is not actually a type, give a meaningful error
+            if not isinstance(ctyp, type):
+                raise TypeError(
+                    "%s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
+                    % (cls.__name__, m, ctyp)
+                )
+
+            component = self._create_component(m, ctyp)
+
+            # Store for later
+            components.append((m, component))
+
+        # - Iterate over set class variables
+        for m in self.members:
+            if m.startswith("_") or isinstance(getattr(cls, m, None), _TunableProperty):
+                continue
+
             ctyp = getattr(self, m)
             if not isinstance(ctyp, type):
                 continue
 
-            # Create instance, set it on self
-            component = ctyp()
-            setattr(self, m, component)
-
-            # Ensure that mandatory methods are there
-            if not hasattr(component, 'execute') or \
-               not callable(component.execute):
-                raise ValueError("Component %s (%s) must have a method named 'execute'" % (m, component))
-
-            # Automatically inject a logger object
-            component.logger = logging.getLogger(m)
-            component._Magicbot__autosend = {}
+            component = self._create_component(m, ctyp)
 
             # Store for later
             components.append((m, component))
-            
-            self.logger.info("-> %s (class: %s)", m, ctyp.__name__)
 
         # Collect attributes of this robot that are injectable
         self._injectables = {}
 
         for n in dir(self):
-            if n.startswith('_') or n in self._exclude_from_injection or \
-               isinstance(getattr(cls, n, None), _TunableProperty):
+            if (
+                n.startswith("_")
+                or n in self._exclude_from_injection
+                or isinstance(getattr(cls, n, None), _TunableProperty)
+            ):
                 continue
 
             o = getattr(self, n)
@@ -383,63 +490,151 @@ class MagicRobot(wpilib.SampleRobot,
         # For each new component, perform magic injection
         for cname, component in components:
             self._components.append(component)
-            setup_tunables(component, cname, 'components')
+            setup_tunables(component, cname, "components")
+            self._feedbacks += collect_feedbacks(component, cname, "components")
             self._setup_vars(cname, component)
+            self._setup_reset_vars(component)
 
         # Do it for autonomous modes too
         for mode in self._automodes.modes.values():
             mode.logger = logging.getLogger(mode.MODE_NAME)
-            setup_tunables(mode, mode.MODE_NAME, 'autonomous')
+            setup_tunables(mode, mode.MODE_NAME, "autonomous")
             self._setup_vars(mode.MODE_NAME, mode)
 
         # And for self too
-        setup_tunables(self, 'robot', None)
-            
+        setup_tunables(self, "robot", None)
+        self._feedbacks += collect_feedbacks(self, "robot", None)
+
         # Call setup functions for components
         for cname, component in components:
-            if hasattr(component, 'setup'):
+            if hasattr(component, "setup"):
                 component.setup()
-    
+
+        # Call setup functions for autonomous modes
+        for mode in self._automodes.modes.values():
+            if hasattr(mode, "setup"):
+                mode.setup()
+
+    def _create_component(self, name, ctyp):
+        # Create instance, set it on self
+        component = ctyp()
+        setattr(self, name, component)
+
+        # Ensure that mandatory methods are there
+        if not callable(getattr(component, "execute", None)):
+            raise ValueError(
+                "Component %s (%r) must have a method named 'execute'"
+                % (name, component)
+            )
+
+        # Automatically inject a logger object
+        component.logger = logging.getLogger(name)
+
+        self.logger.info("-> %s (class: %s)", name, ctyp.__name__)
+
+        return component
+
     def _setup_vars(self, cname, component):
-        
+
         self.logger.debug("Injecting magic variables into %s", cname)
-        
-        for n in dir(component):
-            if n.startswith('_'):
+
+        component_type = type(component)
+
+        # Iterate over variables with type annotations
+        for n, inject_type in get_class_annotations(component_type).items():
+
+            # If the variable is private ignore it
+            if n.startswith("_"):
                 continue
-            
+
+            if hasattr(component_type, n):
+                attr = getattr(component_type, n)
+                # If the value given to the variable is an instance of a type and isn't a property
+                # raise an error. No double declaring types, e.g foo: type = type
+                if isinstance(attr, type) and not isinstance(attr, property):
+                    raise MagicInjectError(
+                        "%s.%s has two type declarations" % (component_type.__name__, n)
+                    )
+                # Otherwise, skip this set class variable
+                continue
+
+            # If the variable has been assigned in __init__, skip it
+            if hasattr(component, n):
+                continue
+
+            # If the type is not actually a type, give a meaningful error
+            if not isinstance(inject_type, type):
+                raise TypeError(
+                    "Component %s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
+                    % (cname, n, inject_type)
+                )
+
+            self._inject(n, inject_type, cname, component)
+
+        # Iterate over static variables
+        for n in dir(component):
+            # If the variable is private or a proprty, don't inject
+            if n.startswith("_") or isinstance(
+                getattr(component_type, n, True), property
+            ):
+                continue
+
             inject_type = getattr(component, n)
-            
+
+            # If the value assigned isn't a type, don't inject
             if not isinstance(inject_type, type):
                 continue
 
-            injectable = self._injectables.get(n)
-            if injectable is None:
-                if cname is not None:
-                    injectable = self._injectables.get('%s_%s' % (cname, n))
+            self._inject(n, inject_type, cname, component)
 
-            if injectable is None:
-                raise ValueError("Component %s has variable %s (type %s), which is not present in %s" %
-                                 (cname, n, inject_type, self))
+    def _inject(self, n, inject_type, cname, component):
+        # Retrieve injectable object
+        injectable = self._injectables.get(n)
+        if injectable is None:
+            if cname is not None:
+                # Try for mangled names
+                injectable = self._injectables.get("%s_%s" % (cname, n))
 
-            if not isinstance(injectable, inject_type):
-                raise ValueError("Component %s variable %s in %s are not the same types! (Got %s, expected %s)" %
-                                 (cname, n, self, type(injectable), inject_type))
+        # Raise error if injectable syntax used but no injectable was found.
+        if injectable is None:
+            raise MagicInjectError(
+                "Component %s has variable %s (type %s), which is not present in %s"
+                % (cname, n, inject_type, self)
+            )
 
-            setattr(component, n, injectable)
-            self.logger.debug("-> %s as %s.%s", injectable, cname, n)
-        
-        # XXX
-        #if is_autosend:
-            # where to store the nt key?
-        #    component._Magicbot__autosend[prop.f] = None
-    
-    #def _update_autosend(self):
-    #    # seems like this should just be a giant list instead
-    #    for component in self._components:
-    #        d = component._Magicbot__autosend
-    #        for f in d.keys():
-    #            d[f] = f(component)     
+        # Raise error if injectable declared with type different than the initial type
+        if not isinstance(injectable, inject_type):
+            raise MagicInjectError(
+                "Component %s variable %s in %s are not the same types! (Got %s, expected %s)"
+                % (cname, n, self, type(injectable), inject_type)
+            )
+        # Perform injection
+        setattr(component, n, injectable)
+        self.logger.debug("-> %s as %s.%s", injectable, cname, n)
+
+    def _setup_reset_vars(self, component):
+        reset_dict = {}
+
+        for n in dir(component):
+            if isinstance(getattr(type(component), n, True), property):
+                continue
+
+            a = getattr(component, n, None)
+            if isinstance(a, will_reset_to):
+                reset_dict[n] = a.default
+
+        if reset_dict:
+            component.__dict__.update(reset_dict)
+            self._reset_components.append((reset_dict, component))
+
+    def _update_feedback(self):
+        for method, entry in self._feedbacks:
+            try:
+                value = method()
+            except:
+                self.onException()
+                continue
+            entry.setValue(value)
 
     def _execute_components(self):
         for component in self._components:
@@ -447,3 +642,6 @@ class MagicRobot(wpilib.SampleRobot,
                 component.execute()
             except:
                 self.onException()
+
+        for reset_dict, component in self._reset_components:
+            component.__dict__.update(reset_dict)
